from .backend.dialer import send_template as send_dialer_template
from .backend.dialer import send as send_dialer

from .backend.emails import send_template as send_email_template
from .backend.emails import send as send_email

from .backend.sms import send_template as send_sms_template
from .backend.sms import send as send_sms


__all__ = (
    'send_dialer_template',
    'send_dialer',
    'send_email_template',
    'send_email',
    'send_sms_template',
    'send_sms',
)



# Generated by Django 3.1 on 2022-01-07 14:42

from django.db import migrations, models
import django.db.models.deletion
import pymess.models.emails


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0025_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='attachment',
            name='filename',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='filename'),
        ),
        migrations.CreateModel(
            name='EmailTemplateAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('content_type', models.CharField(max_length=100, verbose_name='content type')),
                ('file', models.FileField(upload_to=pymess.models.emails.generate_template_attachment_filename,
                                          verbose_name='file')),
                ('filename', models.CharField(blank=True, max_length=100, null=True, verbose_name='filename')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                               related_name='template_attachments', to='pymess.emailtemplate',
                                               verbose_name='template')),
            ],
            options={
                'verbose_name': 'e-mail template attachment',
                'verbose_name_plural': 'e-mail template attachments',
            },
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('pymess', '0028_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailmessage',
            name='pre_header',
            field=models.CharField(blank=True, null=True, max_length=100, verbose_name='pre header'),
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='pre_header',
            field=models.CharField(blank=True, null=True, max_length=100, verbose_name='pre header'),
        ),
    ]



from django.utils import timezone

from pymess.backend.dialer import DialerBackend
from pymess.models import DialerMessage


class DummyDialerBackend(DialerBackend):
    """
    Dummy dialer backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self._update_message_after_sending(message, state=DialerMessage.State.DEBUG, sent_at=timezone.now(),
                                           is_final_state=True)



from django.utils.module_loading import import_string
from pathlib import Path
from uuid import uuid4

from chamber.models import SmartModel

from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.utils.functional import cached_property
from django.db import models
from django.utils.translation import gettext, gettext_lazy as _
from django.template import Template, Context
from django.template.exceptions import TemplateSyntaxError, TemplateDoesNotExist

from pymess.config import settings
from pymess.enums import EmailMessageState
from pymess.utils.html import raise_error_if_contains_banned_tags

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject, MessageQueryset


__all__ = (
    'EmailMessage',
    'EmailRelatedObject',
    'EmailTemplate',
    'Attachment',
    'AbstractEmailTemplate',
    'EmailTemplateDisallowedObject',
    'EmailTemplateAttachment',
)


def generate_content_filename(instance, filename):
    return Path(settings.EMAIL_STORAGE_PATH) / 'contents' / 'content_{}.txt'.format(uuid4())


def generate_attachment_filename(instance, filename):
    return Path(settings.EMAIL_STORAGE_PATH) / 'attachments' / filename


def generate_template_attachment_filename(instance, filename):
    return Path(settings.EMAIL_STORAGE_PATH) / 'template_attachments' / filename


class EmailMessageQuerySet(MessageQueryset):

    def create(self, content, **kwargs):
        message = self.model(**kwargs)
        message.content_file.save(None, ContentFile(content.encode()))
        return message


class EmailMessage(BaseMessage):

    State = EmailMessageState

    recipient = models.EmailField(verbose_name=_('recipient'), blank=False, null=False, db_index=True)
    template = models.ForeignKey(settings.EMAIL_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='email_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False,
                                choices=EmailMessageState.choices, editable=False,
                                db_index=True)
    sender = models.EmailField(verbose_name=_('sender'), blank=False, null=False)
    sender_name = models.CharField(verbose_name=_('sender name'), blank=True, null=True, max_length=250)
    subject = models.TextField(verbose_name=_('subject'), blank=False, null=False)
    pre_header = models.CharField(verbose_name=_('pre header'), blank=True, null=True, max_length=100)
    external_id = models.CharField(verbose_name=_('external ID'), blank=True, null=True, db_index=True, max_length=250)
    last_webhook_received_at = models.DateTimeField(
        verbose_name=_('last webhook received at'),
        null=True,
        blank=True,
        editable=False,
    )
    info_changed_at = models.DateTimeField(
        verbose_name=_('info changed at'),
        null=True,
        blank=True,
        editable=False,
    )
    content_file = models.FileField(
        verbose_name=_('content file'),
        null=False,
        blank=False,
        upload_to=generate_content_filename,
    )

    objects = EmailMessageQuerySet.as_manager()

    class Meta(BaseMessage.Meta):
        verbose_name = _('e-mail message')
        verbose_name_plural = _('e-mail messages')

    def __str__(self):
        return '{}: {}'.format(self.recipient, self.subject)

    @property
    def friendly_sender(self):
        """
        returns sender with sender name in standard address format if sender name was defined
        """
        return '{} <{}>'.format(self.sender_name, self.sender) if self.sender_name else self.sender

    @property
    def failed(self):
        return self.state in {EmailMessageState.ERROR, EmailMessageState.ERROR_RETRY}

    @cached_property
    def content(self):
        return self.content_file.read().decode()


class EmailRelatedObject(BaseRelatedObject):

    email_message = models.ForeignKey(EmailMessage, verbose_name=_('e-mail message'), null=False, blank=False,
                                      on_delete=models.CASCADE, related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a e-mail message')
        verbose_name_plural = _('related objects of e-mail messages')


class AttachmentManager(models.Manager):

    def create_from_tripple(self, tripple):
        filename, file, content_type = tripple
        attachment = self.model(
            email_message=self.instance,
            content_type=content_type,
            filename=filename
        )
        attachment.file.save(filename, file, save=True)
        return attachment

    def create_from_tripples(self, *tripples):
        return [
            self.create_from_tripple(tripple) for tripple in tripples
        ]


class Attachment(SmartModel):

    email_message = models.ForeignKey(EmailMessage, verbose_name=_('e-mail message'), on_delete=models.CASCADE,
                                      related_name='attachments')
    content_type = models.CharField(verbose_name=_('content type'), blank=False, null=False, max_length=100)
    file = models.FileField(verbose_name=_('file'), null=False, blank=False,
                            upload_to=generate_attachment_filename)
    filename = models.CharField(verbose_name=_('filename'), blank=True, null=True, max_length=100)

    objects = AttachmentManager()

    def __str__(self):
        return '#{}'.format(self.pk)

    class Meta:
        verbose_name = _('attachment')
        verbose_name_plural = _('attachments')


class AbstractEmailTemplate(BaseAbstractTemplate):

    subject = models.TextField(verbose_name=_('subject'), blank=False, null=False)
    pre_header = models.CharField(verbose_name=_('pre header'), blank=True, null=True, max_length=100)
    sender = models.EmailField(verbose_name=_('sender'), null=True, blank=True, max_length=200)
    sender_name = models.CharField(verbose_name=_('sender name'), blank=True, null=True, max_length=250)
    variant = models.CharField(verbose_name=_('variant'), max_length=10, null=True, blank=True, editable=False)
    version = models.CharField(verbose_name=_('version'), max_length=10, null=True, blank=True)

    def get_controller(self):
        from pymess.backend.emails import EmailController
        return EmailController()

    def _update_context_data(self, context_data, recipient):
        for context_processor_fun_name in settings.EMAIL_TEMPLATE_CONTEXT_PROCESSORS:
            context_data.update(import_string(context_processor_fun_name)(context_data, self, recipient))
        return context_data

    def clean_subject(self, context_data=None):
        try:
            self.render_subject(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(gettext('Error during template subject rendering: "{}"').format(ex))

    def get_subject(self):
        return self.subject

    def render_subject(self, context_data, recipient=None):
        context_data = self._update_context_data(context_data, recipient)
        return Template(self.get_subject()).render(Context(context_data))

    def send(self, recipient, context_data, related_objects=None, tag=None, attachments=None,
             priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        return super().send(
            recipient=recipient,
            context_data=context_data,
            related_objects=related_objects,
            tag=tag,
            sender=self.sender,
            subject=self.render_subject(context_data),
            pre_header=self.pre_header,
            sender_name=self.sender_name,
            priority=priority,
            attachments=attachments,
            **kwargs,
        )


    def __str__(self):
        return f"{super().__str__()} ({self.variant})" if self.variant else super().__str__()

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('e-mail template')
        unique_together = ("slug", "locale", "variant")
        verbose_name_plural = _('e-mail templates')


class EmailTemplate(AbstractEmailTemplate):

    def clean_body(self, context_data=None):
        super().clean_body(context_data={'EMAIL_DISABLE_VARIABLE_VALIDATOR': True})
        raise_error_if_contains_banned_tags(self.body)

    def clean_subject(self, context_data=None):
        super().clean_subject(context_data={'EMAIL_DISABLE_VARIABLE_VALIDATOR': True})
        raise_error_if_contains_banned_tags(self.subject)

    def get_base_template(self):
        base_template_config = settings.EMAIL_TEMPLATE_BASE_TEMPLATE

        if not base_template_config:
            raise ValidationError(gettext('EMAIL_TEMPLATE_BASE_TEMPLATE is not set'))

        if isinstance(base_template_config, dict):
            if not self.version:
                return next(iter(base_template_config.values()), None)

            return base_template_config.get(self.version)

        return base_template_config


    def _extend_body(self, template_body):
        base_template = self.get_base_template()
        templatetags = settings.EMAIL_TEMPLATE_TEMPLATETAGS

        template_content = '{{% block {} %}}{}{{% endblock %}}'.format(
            settings.EMAIL_TEMPLATE_CONTENT_BLOCK,
            template_body
        )

        out = []
        if base_template is not None:
            out.append('{{% extends \'{}\' %}}'.format(base_template))
        if templatetags:
            out.append('{{% load {} %}}'.format(' '.join(templatetags)))
        out.append(template_content)

        return ''.join(out)

    def get_body(self):
        return self._extend_body(self.body) if settings.EMAIL_TEMPLATE_EXTEND_BODY else self.body

    def send(self, recipient, context_data, related_objects=None, tag=None, attachments=None, **kwargs):
        attachments = [] if attachments is None else attachments
        attachments += [
            (
                template_attachment.filename or os.path.basename(template_attachment.file.name),
                ContentFile(template_attachment.file.read()),
                template_attachment.content_type,
            ) for template_attachment in self.template_attachments.all()
        ]

        return super().send(
            recipient=recipient,
            context_data=context_data,
            related_objects=related_objects,
            tag=tag,
            attachments=attachments,
            **kwargs,
        )


class EmailTemplateDisallowedObject(BaseRelatedObject):

    template = models.ForeignKey(
        verbose_name=_('template'),
        to=EmailTemplate,
        null=True, blank=True,
        on_delete=models.CASCADE,
        related_name='disallowed_objects',
        db_index=True
    )

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('disallowed object of an e-mail template')
        verbose_name_plural = _('disallowed objects of e-mail templates')


class EmailTemplateAttachment(SmartModel):

    template = models.ForeignKey(to=settings.EMAIL_TEMPLATE_MODEL, verbose_name=_('template'),
                                 on_delete=models.CASCADE, related_name='template_attachments')
    content_type = models.CharField(verbose_name=_('content type'), blank=False, null=False, max_length=100)
    file = models.FileField(verbose_name=_('file'), null=False, blank=False,
                            upload_to=generate_template_attachment_filename)
    filename = models.CharField(verbose_name=_('filename'), blank=True, null=True, max_length=100)

    class Meta:
        verbose_name = _('e-mail template attachment')
        verbose_name_plural = _('e-mail template attachments')



from unittest.mock import Mock, patch, call

import pytest
import xml.etree.ElementTree as ET

from pymess.backend.sms.sms_operator import RequestType, SMSOperatorBackend, SmsOperatorState
from pymess.enums import OutputSMSMessageState
from pymess.models.sms import OutputSMSMessage


@pytest.fixture
def backend():
    return SMSOperatorBackend(config={
        'USERNAME': 'user',
        'PASSWORD': 'pass',
        'UNIQ_PREFIX': 'pref',
        'SMS_URL': 'https://sms.example',
        'VOICE_URL': 'https://voice.example',
        'TIMEOUT': 1,
    })


@pytest.mark.django_db
class TestSMSOperatorBackend:
    def test_serialize_voice_message_should_use_voice_template(self, backend):
        message = OutputSMSMessage.objects.create(
            recipient='+420111111111',
            content='Voice content',
            state=OutputSMSMessageState.WAITING,
            is_voice_message=True,
            sender='CallerID',
        )

        xml = backend._serialize_messages([message], RequestType.VOICE_MESSAGE)
        root = ET.fromstring(xml)

        assert root.tag == 'VoiceServices'
        header = root.find('./DataHeader/DataType')
        assert header.text == 'VoiceMessage'
        item = root.find('.//DataItem')
        assert item.find('PhoneNumber').text == '+420111111111'
        assert item.find('./VoiceMsg/Text').text == 'Voice content'
        assert item.find('MsgId').text.endswith(str(message.pk))

    def test_serialize_sms_message_should_use_sms_template(self, backend):
        message = OutputSMSMessage.objects.create(
            recipient='+420222222222',
            content='Sms content',
            state=OutputSMSMessageState.WAITING,
        )

        xml = backend._serialize_messages([message], RequestType.SMS)
        root = ET.fromstring(xml)

        assert root.tag == 'SmsServices'
        header = root.find('./DataHeader/DataType')
        assert header.text == 'SMS'
        item = root.find('.//DataItem')
        assert item.find('SmsId').text.endswith(str(message.pk))

    @pytest.mark.parametrize(
        'request_type,expected_url_attr',
        [
            (RequestType.SMS, 'SMS_URL'),
            (RequestType.VOICE_MESSAGE, 'VOICE_URL'),
        ],
    )
    @patch('pymess.backend.sms.sms_operator.generate_session')
    def test_send_requests_should_use_expected_endpoint(self, mock_generate_session, backend, request_type, expected_url_attr):
        message = OutputSMSMessage.objects.create(
            recipient='+420123456789',
            content='content',
            state=OutputSMSMessageState.WAITING,
            is_voice_message=request_type == RequestType.VOICE_MESSAGE,
        )

        session = Mock()
        if request_type == RequestType.VOICE_MESSAGE:
            body = f'<VoiceServices><DataItem><MsgId>pref-{message.pk}</MsgId><Status>0</Status></DataItem></VoiceServices>'
        else:
            body = f'<SmsServices><DataItem><SmsId>pref-{message.pk}</SmsId><Status>0</Status></DataItem></SmsServices>'
        response = Mock(status_code=200, text=body)
        session.post.return_value = response
        mock_generate_session.return_value = session

        backend._send_requests([message], request_type=request_type, is_sending=True)

        session.post.assert_called_once()
        called_url = session.post.call_args[0][0]
        assert called_url == backend.config[expected_url_attr]

    @patch('pymess.backend.sms.sms_operator.generate_session')
    def test_publish_messages_should_use_endpoint_based_on_sms_message_type(self, mock_generate_session, backend):
        sms = OutputSMSMessage.objects.create(
            recipient='+420111111111',
            content='sms',
            state=OutputSMSMessageState.WAITING,
            is_voice_message=False,
        )
        voice = OutputSMSMessage.objects.create(
            recipient='+420222222222',
            content='voice',
            state=OutputSMSMessageState.WAITING,
            is_voice_message=True,
        )

        session = Mock()
        voice_response = Mock(
            status_code=200,
            text=(
                f'<VoiceServices>'
                f'<DataItem><MsgId>pref-{voice.pk}</MsgId><Status>0</Status></DataItem>'
                f'</VoiceServices>'
            )
        )
        sms_response = Mock(
            status_code=200,
            text=(
                f'<SmsServices>'
                f'<DataItem><SmsId>pref-{sms.pk}</SmsId><Status>0</Status></DataItem>'
                f'</SmsServices>'
            )
        )
        session.post.side_effect = [voice_response, sms_response]
        mock_generate_session.return_value = session

        backend.publish_messages([sms, voice])

        assert session.post.call_count == 2
        called_urls = [call.args[0] for call in session.post.call_args_list]
        assert called_urls[0] == backend.config['VOICE_URL']
        assert called_urls[1] == backend.config['SMS_URL']

    @pytest.mark.parametrize(
        'is_voice_message,expected_request_type,expected_url_key',
        [
            (True, RequestType.VOICE_MESSAGE, 'VOICE_URL'),
            (False, RequestType.SMS, 'SMS_URL'),
        ],
    )
    @patch('pymess.backend.sms.sms_operator.generate_session')
    def test_publish_message_should_use_expected_endpoint(
        self,
        mock_generate_session,
        backend,
        is_voice_message,
        expected_request_type,
        expected_url_key,
    ):
        content = 'voice' if is_voice_message else 'sms'
        message = OutputSMSMessage.objects.create(
            recipient='+420999999999',
            content=content,
            state=OutputSMSMessageState.WAITING,
            is_voice_message=is_voice_message,
        )

        session = Mock()
        template_tag = 'VoiceServices' if is_voice_message else 'SmsServices'
        id_tag = 'MsgId' if is_voice_message else 'SmsId'
        body = (
            f'<{template_tag}>'
            f'<DataItem><{id_tag}>pref-{message.pk}</{id_tag}><Status>0</Status></DataItem>'
            f'</{template_tag}>'
        )
        response = Mock(status_code=200, text=body)
        session.post.return_value = response
        mock_generate_session.return_value = session

        with patch.object(backend, '_send_requests', wraps=backend._send_requests) as mock_send_requests:
            backend.publish_message(message)

        mock_send_requests.assert_called_once()
        send_args, send_kwargs = mock_send_requests.call_args
        assert send_args[0] == [message]
        assert send_kwargs['request_type'] == expected_request_type
        assert send_kwargs['is_sending'] is True
        assert 'sent_at' in send_kwargs

        called_url = session.post.call_args[0][0]
        assert called_url == backend.config[expected_url_key]

    def test_update_sms_states_should_send_requests_per_message_type(self, backend):
        sms = OutputSMSMessage.objects.create(
            recipient='+420111111111',
            content='sms',
            state=OutputSMSMessageState.SENDING,
            is_voice_message=False,
        )
        voice = OutputSMSMessage.objects.create(
            recipient='+420222222222',
            content='voice',
            state=OutputSMSMessageState.SENDING,
            is_voice_message=True,
        )

        with patch.object(backend, '_send_requests') as mock_send_requests:
            backend.update_sms_states([sms, voice])

        expected_calls = [
            call([voice], request_type=RequestType.VOICE_MESSAGE_DELIVERY_REQUEST),
            call([sms], request_type=RequestType.DELIVERY_REQUEST),
        ]
        mock_send_requests.assert_has_calls(expected_calls)
        assert mock_send_requests.call_count == 2

    def test_parse_response_codes_should_accept_msgid_and_smsid(self, backend):
        xml = '''
            <VoiceServices>
              <DataItem><MsgId>pref-10</MsgId><Status>0</Status></DataItem>
              <DataItem><SmsId>pref-11</SmsId><Status>1</Status></DataItem>
            </VoiceServices>
        '''

        result = backend._parse_response_codes(xml)

        assert result[10] == SmsOperatorState.DELIVERED
        assert result[11] == SmsOperatorState.NOT_DELIVERED



# Generated by Django 4.2.18 on 2025-01-16 15:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0031_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='smstemplate',
            name='sender_name',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
    ]



import boto3

from django.conf import settings
from django.utils import timezone

from pymess.config import settings
from pymess.backend.sms import SMSBackend
from pymess.enums import OutputSMSMessageState


class SNSSMSBackend(SMSBackend):
    """
    SMS backend implementing AWS SNS service via boto3 library https://aws.amazon.com/sns/
    """

    sns_client = None
    config = {
        'AWS_ACCESS_KEY_ID': None,
        'AWS_SECRET_ACCESS_KEY': None,
        'AWS_REGION': None,
        'SENDER_ID': None,
    }

    def _get_sns_client(self):
        """
        Connect to the SNS service
        """
        if not self.sns_client:
            self.sns_client = boto3.client(
                service_name='sns',
                aws_access_key_id=self.config['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=self.config['AWS_SECRET_ACCESS_KEY'],
                region_name=self.config['AWS_REGION'],
                use_ssl=True
            )
        return self.sns_client

    def publish_message(self, message):
        """
        Method uses boto3 client via witch SMS message is send
        :param message: SMS message
        """
        sns_client = self._get_sns_client()
        publish_kwargs = {
            'PhoneNumber': str(message.recipient),
            'Message': message.content,
        }
        if self.config['SENDER_ID']:
            publish_kwargs.update({
                'MessageAttributes': {
                    'AWS.SNS.SMS.SenderID': {
                        'DataType': 'String',
                        'StringValue': self.config['SENDER_ID'],
                    }
                }
            })
        try:
            sns_client.publish(**publish_kwargs)
            self._update_message_after_sending(message, state=OutputSMSMessageState.SENT, sent_at=timezone.now())
        except Exception as ex:
            self._update_message_after_sending_error(
                message, error=str(ex)
            )
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).



import pytest
from unittest.mock import Mock, patch

from pymess.backend.sms import SMSController
from pymess.enums import OutputSMSMessageState
from pymess.models.sms import OutputSMSMessage, SMSTemplate


@pytest.mark.django_db
class TestSMSController:
    def test_create_message_should_inherit_voice_flag_from_template(self):
        template = SMSTemplate.objects.create(
            slug='voice-otp',
            body='Voice body',
            sender_name='MyApp',
            is_voice_message=True,
        )
        controller = SMSController()

        with patch.object(controller, 'get_backend') as mock_get_backend:
            backend = Mock()
            backend.get_extra_message_kwargs.return_value = {}
            mock_get_backend.return_value = backend

            message = controller.create_message(
                recipient='+420123456789',
                content='content',
                related_objects=[],
                tag='otp',
                template=template,
            )

        assert message.is_voice_message is True
        assert message.sender == 'MyApp'
        mock_get_backend.assert_called_once_with('+420123456789', is_voice_message=True)

    def test_create_message_without_template_should_default_to_sms(self):
        controller = SMSController()

        with patch.object(controller, 'get_backend') as mock_get_backend:
            backend = Mock()
            backend.get_extra_message_kwargs.return_value = {}
            mock_get_backend.return_value = backend

            message = controller.create_message(
                recipient='+420123456789',
                content='content',
                related_objects=[],
                tag='otp',
                template=None,
            )

        assert message.is_voice_message is False
        mock_get_backend.assert_called_once_with('+420123456789', is_voice_message=False)

    def test_get_backend_should_pass_voice_flag_to_router(self):
        controller = SMSController()

        with patch.object(controller, 'router') as mock_router:
            with patch('pymess.backend.get_backend') as mock_get_backend:
                backend = Mock()
                backend.get_extra_message_kwargs.return_value = {}
                mock_get_backend.return_value = backend
                mock_router.get_backend_name.return_value = 'dummy-backend'

                controller.get_backend(recipient='+420123456789', is_voice_message=True)

        mock_router.get_backend_name.assert_called_once_with('+420123456789', is_voice_message=True)
        mock_get_backend.assert_called_once_with(controller.backend_type_name, 'dummy-backend')



from chamber.utils import remove_accent
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from pymess.config import settings
from pymess.utils import normalize_phone_number
from pymess.enums import OutputSMSMessageState

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject


__all__ = (
    'OutputSMSMessage',
    'OutputSMSRelatedObject',
    'AbstractSMSTemplate',
    'SMSTemplate',
    'SMSTemplateDisallowedObject',
)


class OutputSMSMessage(BaseMessage):

    State = OutputSMSMessageState

    content = models.TextField(verbose_name=_('content'), null=False, blank=False, max_length=700)
    template = models.ForeignKey(settings.SMS_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='output_sms_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False,
                                choices=OutputSMSMessageState.choices, editable=False,
                                db_index=True)
    sender = models.CharField(verbose_name=_('sender'), null=True, blank=True, max_length=20)
    is_voice_message = models.BooleanField(verbose_name=_('is voice message'), null=False, blank=False,
                                           default=False)

    class Meta(BaseMessage.Meta):
        verbose_name = _('output SMS')
        verbose_name_plural = _('output SMS')

    def clean_recipient(self):
        self.recipient = normalize_phone_number(force_str(self.recipient))

    def clean_content(self):
        if not settings.SMS_USE_ACCENT:
            self.content = str(remove_accent(str(self.content)))

    @property
    def failed(self):
        return self.state in {OutputSMSMessageState.ERROR, OutputSMSMessageState.ERROR_RETRY}


class OutputSMSRelatedObject(BaseRelatedObject):

    output_sms_message = models.ForeignKey(OutputSMSMessage, verbose_name=_('output SMS message'), null=False,
                                           blank=False, on_delete=models.CASCADE, related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a SMS message')
        verbose_name_plural = _('related objects of SMS messages')


class AbstractSMSTemplate(BaseAbstractTemplate):

    is_secret = models.BooleanField(null=False, blank=False, default=True, verbose_name=_('is secret'))
    sender_name = models.CharField(null=True, blank=True, max_length=20, verbose_name=_('sender name'))
    is_voice_message = models.BooleanField(verbose_name=_('is voice message'), null=False, blank=False,
                                           default=False)

    def get_controller(self):
        from pymess.backend.sms import SMSController
        return SMSController()

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('SMS template')
        verbose_name_plural = _('SMS templates')


class SMSTemplate(AbstractSMSTemplate):
    pass


class SMSTemplateDisallowedObject(BaseRelatedObject):

    template = models.ForeignKey(
        verbose_name=_('template'),
        to=SMSTemplate,
        null=True, blank=True,
        on_delete=models.CASCADE,
        related_name='disallowed_objects',
        db_index=True
    )

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('disallowed object of an SMS template')
        verbose_name_plural = _('disallowed objects of SMS templates')



# Generated by Django 4.2.25 on 2025-10-12 20:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0033_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='outputsmsmessage',
            name='is_voice_message',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='smstemplate',
            name='is_voice_message',
            field=models.BooleanField(default=False),
        ),
    ]



import os

from django.core.mail import EmailMultiAlternatives
from django.utils import timezone

from pymess.backend.emails import EmailBackend


class SMTPEmailBackend(EmailBackend):
    """
    E-mail backend implementing standard SMTP service
    """

    def publish_message(self, message):
        email_message = EmailMultiAlternatives(
            message.subject,
            ' ',
            message.friendly_sender,
            [message.recipient],
        )
        email_message.attach_alternative(message.content, 'text/html')
        for attachment in message.attachments.all():
            email_message.attach(
                attachment.filename or os.path.basename(attachment.file.name),
                attachment.file.read().decode('utf-8'),
                attachment.content_type,
            )
        try:
            email_message.send()
            self._update_message_after_sending(message, state=message.State.SENT, sent_at=timezone.now())
        except Exception as ex:
            self._update_message_after_sending_error(message, error=str(ex))
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).



import codecs
import os

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand

from pymess.config import get_email_template_model, settings


class Command(BaseCommand):
    """
    Command to synchronize e-mail bodies from directory.
    Default templates remain the same. Variant B templates have the '_B' suffix.
    """

    def _sync_for_locale(self, base_directory, locale, variant):
        locale_directory = os.path.join(base_directory, locale) if locale else base_directory

        files_with_email_html_body = [
            filename for filename in os.listdir(locale_directory) if filename.endswith('.html')
        ]
        for filename in files_with_email_html_body:
            slug, _ = filename.rsplit('_' if variant else '.html', 1)
            try:
                template_obj = get_email_template_model().objects.get(slug=slug, locale=locale, variant=variant)
                with codecs.open(os.path.join(locale_directory, filename), 'r', encoding='utf-8-sig') as file:
                    template_obj.change_and_save(body=file.read())
            except ObjectDoesNotExist:
                if verbosity > 0:
                    self.stderr.write('Template model with slug "{}" does not exists'.format(filename[:-5]))
        if verbosity > 0:
            self.stdout.write(
                self.style.SUCCESS('Synced "{}" e-mail templates'.format(len(files_with_email_html_body)))
            )

    def handle(self, *args, **options):
        verbosity = int(options.get('verbosity'))
        variant = str(options.get('variant'))
        if verbosity > 0:
            self.stdout.write('Syncing e-mails')

        directory = os.path.join(settings.EMAIL_HTML_DATA_DIRECTORY)
        # All dirs in the root directory represent a locale.
        # e.g., `$EMAIL_HTML_DATA_DIRECTORY/cs/`, `$EMAIL_HTML_DATA_DIRECTORY/sk/`.
        for filename in os.listdir(directory):
            file_path = os.path.join(directory, filename)
            if os.path.isdir(file_path):
                self._sync_for_locale(directory, filename, variant)

        # The `$EMAIL_HTML_DATA_DIRECTORY/` directory itself represents the locale of None value.
        self._sync_for_locale(directory, None, variant)



from django.utils import timezone

from pymess.backend.push import PushNotificationBackend
from pymess.models import PushNotificationMessage


class DummyPushNotificationBackend(PushNotificationBackend):
    """Dummy push notification backend used for testing environments. Backend only logs messages to the database."""

    def publish_message(self, message):
        self._update_message_after_sending(message, state=PushNotificationMessage.State.DEBUG, sent_at=timezone.now())



from django.utils.translation import gettext_lazy as _

from enumfields import IntegerChoicesEnum


class DialerMessageState(IntegerChoicesEnum):
     # numbers are matching predefined state values in Daktela

     WAITING = -1, _('waiting')
     NOT_ASSIGNED = 0, _('not assigned')
     READY = 1, _('ready')
     RESCHEDULED_BY_DIALER = 2, _('rescheduled by dialer')
     CALL_IN_PROGRESS = 3, _('call in progress')
     HANGUP = 4, _('hangup')
     DONE = 5, _('done')
     RESCHEDULED = 6, _('rescheduled')
     ANSWERED_COMPLETE = 7, _('listened up complete message')
     ANSWERED_PARTIAL = 8, _('listened up partial message')
     UNREACHABLE = 9, _('unreachable')
     DECLINED = 10, _('declined')
     UNANSWERED = 11, _('unanswered')
     HANGUP_BY_DIALER = 12, _('unanswered - hangup by dialer')
     HANGUP_BY_CUSTOMER = 13, _('answered - hangup by customer')
     ERROR_UPDATE = 66, _('error message update')
     DEBUG = 77, _('debug')
     ERROR = 88, _('error')
     ERROR_RETRY = 99, _('error retry')


class EmailMessageState(IntegerChoicesEnum):

    WAITING = 1, _('waiting')
    SENDING = 2, _('sending')
    SENT = 3, _('sent')
    ERROR = 4, _('error')
    DEBUG = 5, _('debug')
    ERROR_RETRY = 6, _('error retry')


class PushNotificationMessageState(IntegerChoicesEnum):

    WAITING = 1, _('waiting')
    SENT = 2, _('sent')
    ERROR = 3, _('error')
    DEBUG = 4, _('debug')
    ERROR_RETRY = 5, _('error retry')


class OutputSMSMessageState(IntegerChoicesEnum):

    WAITING = 1, _('waiting')
    UNKNOWN = 2, _('unknown')
    SENDING = 3, _('sending')
    SENT = 4, _('sent')
    ERROR_UPDATE = 5, _('error message update')
    DEBUG = 6, _('debug')
    DELIVERED = 7, _('delivered')
    ERROR = 8, _('error')
    ERROR_RETRY = 9, _('error retry')



"""
Django test settings for pymess project.
"""
import os

SECRET_KEY = 'test-key'
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.admin',
    'pymess',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'pymess', 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

USE_TZ = True 


# Generated by Django 4.2.23 on 2025-08-06 10:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0032_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='version',
            field=models.CharField(verbose_name='version', blank=True, max_length=10, null=True),
        ),
    ]



from django.db import migrations, models



class Migration(migrations.Migration):
    dependencies = [
        ('pymess', '0030_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name="emailtemplate",
            name="variant",
            field=models.CharField(blank=True, editable=False, max_length=10, null=True),
        ),
        migrations.AlterUniqueTogether(
            name="emailtemplate",
            unique_together={("slug", "locale", "variant")},
        ),
    ]



import pytest
from pymess.models.emails import EmailMessage
from pymess.enums import EmailMessageState


@pytest.mark.django_db
class TestEmailMessage:

    def test_email_message_failed_property(self):
        """
        Test that the failed property correctly identifies error states.
        """
        # Test non-error state
        message = EmailMessage.objects.create(
            recipient='test@example.com',
            sender='sender@example.com',
            subject='Test Subject',
            content='Test Content',
            state=EmailMessageState.WAITING,
        )
        assert message.failed is False

        # Test error state
        message.change_and_save(state=EmailMessageState.ERROR)
        assert message.failed is True

        # Test error retry state
        message.change_and_save(state=EmailMessageState.ERROR_RETRY)
        assert message.failed is True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-11-20 10:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0009_20191108_2007'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='emailmessage',
            name='require_pull_info',
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='info_changed_at',
            field=models.DateTimeField(blank=True, editable=False, null=True, verbose_name='info changed at'),
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='last_webhook_received_at',
            field=models.DateTimeField(blank=True, editable=False, null=True, verbose_name='last webhook received at'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-07-26 12:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0024_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='dialertemplate',
            name='is_locked',
            field=models.BooleanField(default=False, verbose_name='is locked'),
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='is_locked',
            field=models.BooleanField(default=False, verbose_name='is locked'),
        ),
        migrations.AddField(
            model_name='pushnotificationtemplate',
            name='is_locked',
            field=models.BooleanField(default=False, verbose_name='is locked'),
        ),
        migrations.AddField(
            model_name='smstemplate',
            name='is_locked',
            field=models.BooleanField(default=False, verbose_name='is locked'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-12-10 16:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0010_20191120_1002'),
    ]

    operations = [
        migrations.AddField(
            model_name='dialermessage',
            name='is_autodialer',
            field=models.BooleanField(default=True, verbose_name='is autodialer'),
        ),
        migrations.AlterField(
            model_name='dialermessage',
            name='content',
            field=models.TextField(blank=True, null=True, verbose_name='content'),
        ),
        migrations.AlterField(
            model_name='dialermessage',
            name='state',
            field=models.IntegerField(choices=[(0, 'not assigned'), (1, 'ready'), (2, 'rescheduled by dialer'), (3, 'call in progress'), (4, 'hangup'), (5, 'done'), (6, 'rescheduled'), (7, 'listened up complete message'), (8, 'listened up partial message'), (9, 'unreachable'), (10, 'declined'), (11, 'unanswered'), (12, 'answered - hangup by dialer'), (13, 'answered - hangup by customer'), (66, 'error'), (77, 'debug')], editable=False, verbose_name='state'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-11-08 20:07
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0008_auto_20190726_1459'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailmessage',
            name='require_pull_info',
            field=models.BooleanField(default=False, verbose_name='require to pull message info from email service'),
        ),
    ]



from bs4 import BeautifulSoup

import requests

from enum import Enum

from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string

from enumfields import IntegerChoicesEnum

from pymess.backend.sms import SMSBackend
from pymess.enums import OutputSMSMessageState
from pymess.utils.logged_requests import generate_session


class RequestType(str, Enum):

    SMS = 'SMS'
    DELIVERY_REQUEST = 'DELIVERY_REQUEST'


class AtsState(IntegerChoicesEnum):

    # SMS delivery receipts
    NOT_FOUND = 20, _('not found')
    NOT_SENT = 21, _('not sent yet')
    SENT = 22, _('sent')
    DELIVERED = 23, _('delivered')
    NOT_DELIVERED = 24, _('not delivered')
    UNKNOWN = 25, _('not able to determine the state')
    # Authentication
    AUTHENTICATION_FAILED = 100, _('authentication failed')
    # Internal errors
    DB_ERROR = 200, _('DB error')
    # Request states
    OK = 0, _('SMS is OK and ready to be sent')
    UNSPECIFIED_ERROR = 1, _('unspecified error')
    BATCH_WITH_NOT_UNIQUE_UNIQ = 300, _('one of the requests has not unique "uniq"')
    SMS_NOT_UNIQUE_UNIQ = 310, _('SMS has not unique "uniq"')
    SMS_NO_KW = 320, _('SMS lacks keyword')
    KW_INVALID = 321, _('keyword not valid')
    NO_SENDER = 330, _('no sender specified')
    SENDER_INVALID = 331, _('sender not valid')
    MO_PR_NOT_ALLOWED = 332, _('MO PR SMS not allowed')
    MT_PR_NOT_ALLOWED = 333, _('MT PR SMS not allowed')
    MT_PR_DAILY_LIMIT = 334, _('MT PR SMS daily limit exceeded')
    MT_PR_TOTAL_LIMIT = 335, _('MT PR SMS total limit exceeded')
    GEOGRAPHIC_NOT_ALLOWED = 336, _('geographic number is not allowed')
    MT_SK_NOT_ALLOWED = 337, _('MT SMS to Slovakia not allowed')
    SHORTCODES_NOT_ALLOWED = 338, _('shortcodes not allowed')
    UNKNOWN_SENDER = 339, _('sender is unknown')
    UNSPECIFIED_SMS_TYPE = 340, _('type of SMS not specified')
    TOO_LONG = 341, _('SMS too long'),
    TOO_MANY_PARTS = 342, _('too many SMS parts (max. is 10)')
    WRONG_SENDER_OR_RECEIVER = 343, _('wrong number of sender/receiver')
    NO_RECIPIENT_OR_WRONG_FORMAT = 350, _('recipient is missing or in wrong format')
    TEXTID_NOT_ALLOWED = 360, _('using "textid" is not allowed')
    WRONG_TEXTID = 361, _('"textid" is in wrong format')
    LONG_SMS_TEXTID_NOT_ALLOWED = 362, _('long SMS with "textid" not allowed')
    # XML errors
    XML_MISSING = 701, _('XML body missing')
    XML_UNREADABLE = 702, _('XML is not readable')
    WRONG_HTTP_METHOD = 703, _('unknown HTTP method or not HTTP POST')
    XML_INVALID = 705, _('XML invalid')


class ATSSMSBackend(SMSBackend):
    """
    SMS backend that implements ATS operator service https://www.atspraha.cz/
    Backend supports check SMS delivery
    """

    TEMPLATES = {
        'base': 'pymess/sms/ats/base.xml',
        RequestType.SMS: 'pymess/sms/ats/sms.xml',
        RequestType.DELIVERY_REQUEST: 'pymess/sms/ats/delivery_request.xml',
    }

    class ATSSendingError(Exception):
        pass

    ATS_STATES_MAPPING = {
        AtsState.NOT_FOUND: OutputSMSMessageState.ERROR,
        AtsState.NOT_SENT: OutputSMSMessageState.SENDING,
        AtsState.SENT: OutputSMSMessageState.SENT,
        AtsState.DELIVERED: OutputSMSMessageState.DELIVERED,
        AtsState.NOT_DELIVERED: OutputSMSMessageState.ERROR,
        AtsState.OK: OutputSMSMessageState.SENDING,
        AtsState.UNSPECIFIED_ERROR: OutputSMSMessageState.ERROR,
        AtsState.BATCH_WITH_NOT_UNIQUE_UNIQ: OutputSMSMessageState.ERROR,
        AtsState.SMS_NOT_UNIQUE_UNIQ: OutputSMSMessageState.ERROR,
        AtsState.SMS_NO_KW: OutputSMSMessageState.ERROR,
        AtsState.KW_INVALID: OutputSMSMessageState.ERROR,
        AtsState.NO_SENDER: OutputSMSMessageState.ERROR,
        AtsState.SENDER_INVALID: OutputSMSMessageState.ERROR,
        AtsState.MO_PR_NOT_ALLOWED: OutputSMSMessageState.ERROR,
        AtsState.MT_SK_NOT_ALLOWED: OutputSMSMessageState.ERROR,
        AtsState.SHORTCODES_NOT_ALLOWED: OutputSMSMessageState.ERROR,
        AtsState.UNKNOWN_SENDER: OutputSMSMessageState.ERROR,
        AtsState.UNSPECIFIED_SMS_TYPE: OutputSMSMessageState.ERROR,
        AtsState.TOO_LONG: OutputSMSMessageState.ERROR,
        AtsState.TOO_MANY_PARTS: OutputSMSMessageState.ERROR,
        AtsState.WRONG_SENDER_OR_RECEIVER: OutputSMSMessageState.ERROR,
        AtsState.NO_RECIPIENT_OR_WRONG_FORMAT: OutputSMSMessageState.ERROR,
        AtsState.TEXTID_NOT_ALLOWED: OutputSMSMessageState.ERROR,
        AtsState.WRONG_TEXTID: OutputSMSMessageState.ERROR,
        AtsState.LONG_SMS_TEXTID_NOT_ALLOWED: OutputSMSMessageState.ERROR,
    }

    config = {
        'UNIQ_PREFIX': '',
        'VALIDITY': 60,
        'TEXTID': None,
        'URL': 'http://fik.atspraha.cz/gwfcgi/XMLServerWrapper.fcgi',
        'OPTID': '',
        'TIMEOUT': 5,  # 5s
    }

    def _get_extra_sender_data(self):
        return {
            'prefix': self.config['UNIQ_PREFIX'],
            'validity': self.config['VALIDITY'],
            'kw': self.config['PROJECT_KEYWORD'],
            'textid': self.config['TEXTID'],
        }

    def get_extra_message_kwargs(self):
        return {
            'sender': self.config['OUTPUT_SENDER_NUMBER'],
        }

    def _serialize_messages(self, messages, request_type):
        """
        Serialize SMS messages to the XML
        :param messages: list of SMS messages
        :param request_type: type of the request to the ATS operator
        :return: serialized XML message that will be sent to the ATS service
        """
        return render_to_string(
            self.TEMPLATES['base'], {
                'username': self.config['USERNAME'],
                'password': self.config['PASSWORD'],
                'template_type': self.TEMPLATES[request_type],
                'messages': messages,
                'prefix': str(self.config['UNIQ_PREFIX']) + '-',
                'sender': self.config['OUTPUT_SENDER_NUMBER'],
                'dlr': 1,
                'validity': self.config['VALIDITY'],
                'kw': self.config['PROJECT_KEYWORD'],
                'billing': 0,
                'extra': mark_safe(' textid="{textid}"'.format(
                    textid=self.config['TEXTID']
                )) if self.config['TEXTID'] else '',
            }
        )

    def _send_requests(self, messages, request_type, is_sending=False, **change_sms_kwargs):
        """
        Performs the actual POST request for input messages and request type.
        :param messages: list of SMS messages
        :param request_type: type of the request
        :param is_sending: True if method is called after sending message
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """
        requests_xml = self._serialize_messages(messages, request_type)
        try:
            resp = generate_session(slug='pymess - ATS SMS', related_objects=list(messages)).post(
                self.config['URL'],
                data=requests_xml,
                headers={'Content-Type': 'text/xml'},
                timeout=self.config['TIMEOUT']
            )
            if resp.status_code != 200:
                raise self.ATSSendingError(
                    'ATS operator returned invalid response status code: {}'.format(resp.status_code)
                )
            self._update_sms_states_from_response(
                messages, self._parse_response_codes(resp.text), is_sending, **change_sms_kwargs
            )
        except requests.exceptions.RequestException as ex:
            raise self.ATSSendingError(
                'ATS operator returned returned exception: {}'.format(str(ex))
            )

    def _update_sms_states_from_response(self, messages, parsed_response, is_sending=False, **change_sms_kwargs):
        """
        Higher-level function performing serialization of ATS requests, parsing ATS server response and updating
        SMS messages state according the received response.
        :param messages: list of SMS messages
        :param parsed_response: parsed HTTP response from the ATS service
        :param is_sending: True if update is called after sending message
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """

        messages_dict = {message.pk: message for message in messages}

        missing_uniq = set(messages_dict.keys()) - set(parsed_response.keys())
        if missing_uniq:
            raise self.ATSSendingError(
                'ATS operator not returned SMS info with uniq: {}'.format(', '.join(map(str, missing_uniq)))
            )

        extra_uniq = set(parsed_response.keys()) - set(messages_dict.keys())
        if extra_uniq:
            raise self.ATSSendingError(
                'ATS operator returned SMS info about unknown uniq: {}'.format(', '.join(map(str, extra_uniq)))
            )

        for uniq, ats_state in parsed_response.items():
            sms = messages_dict[uniq]
            state = self.ATS_STATES_MAPPING.get(ats_state)
            error = ats_state.label if state == OutputSMSMessageState.ERROR else None
            if is_sending:
                if error:
                    self._update_message_after_sending_error(
                        sms,
                        state=state,
                        error=error,
                        extra_sender_data={'sender_state': ats_state},
                        **change_sms_kwargs
                    )
                else:
                    self._update_message_after_sending(
                        sms,
                        state=state,
                        extra_sender_data={'sender_state': ats_state},
                        **change_sms_kwargs
                    )
            else:
                self._update_message(
                    sms,
                    state=state,
                    error=error,
                    extra_sender_data={'sender_state': ats_state},
                    **change_sms_kwargs
                )

    def publish_messages(self, messages):
        self._send_requests(messages, request_type=RequestType.SMS, is_sending=True, sent_at=timezone.now())

    def publish_message(self, message):
        try:
            self._send_requests(
                [message],
                request_type=RequestType.SMS,
                is_sending=True,
                sent_at=timezone.now()
            )
        except self.ATSSendingError as ex:
            self._update_message_after_sending_error(
                message,
                state=OutputSMSMessageState.ERROR,
                error=str(ex),
            )
        except requests.exceptions.RequestException as ex:
            # Service is probably unavailable sending will be retried
            self._update_message_after_sending_error(
                message,
                error=str(ex)
            )
            # Do not re-raise caught exception. Re-raise exception causes transaction rollback (lost of information
            # about exception).

    def _parse_response_codes(self, xml):
        """
        Finds all <code> tags in the given XML and returns a mapping "uniq" -> "response code" for all SMS.
        In case of an error, the error is logged.
        :param xml: XML from the ATL response
        :return: dictionary with pair {SMS uniq: response status code}
        """

        soup = BeautifulSoup(xml, 'html.parser')
        code_tags = soup.find_all('code')

        error_messages = []
        for c in [int(error_code.string) for error_code in code_tags if not error_code.attrs.get('uniq')]:
            try:
                error_messages.append(AtsState(c).label)
            except ValueError:
                error_messages.append('ATS returned an unknown state {}.'.format(c))
        error_message = ', '.join(error_messages)

        if error_message:
            raise self.ATSSendingError('Error returned from ATS operator: {}'.format(error_message))

        return {
            int(code.attrs['uniq'].lstrip(str(self.config['UNIQ_PREFIX']) + '-')): AtsState(int(code.string))
            for code in code_tags if code.attrs.get('uniq')
        }

    def update_sms_states(self, messages):
        self._send_requests(messages, request_type=RequestType.DELIVERY_REQUEST)



class BaseRouter:

    def get_backend_name(self, recipient, **kwargs):
        """
        Method should return name of the backend specified in PYMESS_*_BACKEND_ROUTER setting option
        """
        raise NotImplementedError


class DefaultBackendRouter(BaseRouter):

    def get_backend_name(self, recipient, **kwargs):
        return None
